"""
@author: Lowinli
@email: lowinli@outlook.com

"""

from transformers import AutoTokenizer, AutoModelForCausalLM
from onnxruntime.quantization import quantize_dynamic, QuantType
import onnxruntime
import torch
import os
import time
import numpy as np
from torch import tensor
from .model_wrapper import ModelWrapper


def generate_onnx_representation(model, onnx_name_path=None):
    os.makedirs(os.path.join(model.config._name_or_path, "onnx"), exist_ok=True)
    if onnx_name_path is None:
        onnx_name_path = os.path.join(model.config._name_or_path, "onnx", "model.onnx")
    n_layer = model.config.n_layer
    n_head = model.config.n_head
    embed_size_per_head = int(model.config.n_embd / model.config.n_head)
    model_wrapper = ModelWrapper(model)
    model_wrapper.eval()
    past_key_values = torch.randn([n_layer, 2, 1, n_head, 1, embed_size_per_head])
    # past_key_values = torch.randn([n_layer, 2, 1, n_head, int(n_positions/4), embed_size_per_head])
    # 如果解码第一个token，没有past_key_values, 则使用torch.randn([1, 12, 0, 64])，在transformers中等同于past_key_values=None

    test_inputs = {
        "input_ids": torch.LongTensor([[4299, 23748, 62, 6894, 33529]]),
        "attention_mask": torch.LongTensor([[1, 1, 1, 1, 1, 1]]),
        "past_key_values": past_key_values,
    }
    print("onnx dummpy输入配置")
    dummy_inputs = tuple([x for x in test_inputs.values()])
    dynamic_axes = {
        "input_ids": {0: "batch_size", 1: "seq_len"},  # 动态维度,第0和1维度都是可变的
        "attention_mask": {0: "batch_size", 1: "seq_len"},  # 动态维度,第0和1维度都是可变的
        "past_key_values": {
            2: "batch_size",
            4: "seq_len",
        },  # 层数(不可变), QK(不可变), batch_size(可变), num_heads(不可变), sequence_length(可变), embed_size_per_head(不可变)
    }
    print("onnx转换中...")
    # 导出
    torch.onnx.export(
        model_wrapper,  # 模型
        args=dummy_inputs,  # 模型输入
        f=onnx_name_path,  # 存储路径
        export_params=True,
        opset_version=14,
        do_constant_folding=True,
        input_names=[
            "input_ids",
            "attention_mask",
            "past_key_values",
        ],
        # 模型输入的名字，随便写
        output_names=["logits", "past_key_values_array"],  # 模型输出的名字，使用ort就要根据这个名字获取结果
        dynamic_axes=dynamic_axes,
    )
    return onnx_name_path


def quantize(onnx_name_path):
    print("onnx量化int8中...")
    model_name = onnx_name_path
    output_model_name = f"{model_name[:-5]}-quantized.onnx"
    quantize_dynamic(
        model_input=model_name,
        model_output=output_model_name,
        per_channel=True,
        reduce_range=True,  # should be the same as per_channel
        activation_type=QuantType.QUInt8,
        weight_type=QuantType.QInt8,  # per docs, signed is faster on most CPUs
        optimize_model=False,
    )
    return output_model_name


def test_torch_inference(model):
    print("torch加载测试...")
    n_layer = model.config.n_layer
    n_head = model.config.n_head
    embed_size_per_head = int(model.config.n_embd / model.config.n_head)
    model_wrapper = ModelWrapper(model)
    model_wrapper.eval()
    past_key_values = torch.randn([n_layer, 2, 1, n_head, 0, embed_size_per_head])
    # past_key_values = torch.randn([n_layer, 2, 1, n_head, int(n_positions/4), embed_size_per_head])
    # 如果解码第一个token，没有past_key_values, 则使用torch.randn([1, 12, 0, 64])，在transformers中等同于past_key_values=None
    test_inputs = {
        "input_ids": tensor([[4299, 23748, 62, 6894, 33529]]),
        "attention_mask": tensor([[1, 1, 1, 1, 1]]),
        "past_key_values": past_key_values,
    }
    start = time.time()
    with torch.no_grad():
        output = model_wrapper(**test_inputs)
    print("#" * 10 + "\ntorch格式输出：\n")
    print(output[0])
    waste_time = time.time() - start
    latency = round(waste_time * 1000, 3)
    print(f"torch推断用时{latency}ms")
    return output[0]


def test_onnx_inference(onnx_name_path, config):
    print("onnx加载测试...")
    n_layer = config.n_layer
    n_head = config.n_head
    embed_size_per_head = int(config.n_embd / config.n_head)
    session = onnxruntime.InferenceSession(onnx_name_path)
    ort_inputs = {
        "input_ids": np.array([[4299, 23748, 62, 6894, 33529]], dtype=np.int64),
        "attention_mask": np.array([[1, 1, 1, 1, 1]], dtype=np.int64),
        "past_key_values": np.random.normal(
            size=(n_layer, 2, 1, n_head, 0, embed_size_per_head)
        ).astype(np.float32),
    }
    start = time.time()
    output = session.run(None, ort_inputs)
    print("#" * 10 + "\nonnx格式输出：\n")
    print(output[0])
    waste_time = time.time() - start
    latency = round(waste_time * 1000, 3)
    print(f"onnx推断用时{latency}ms")
    return output[0]


def download_from_transformers(model_name_path):
    print("下载torch模型中...")
    tokenizer = AutoTokenizer.from_pretrained(model_name_path)
    model = AutoModelForCausalLM.from_pretrained(model_name_path)
    tokenizer.save_pretrained(model_name_path)
    model.save_pretrained(model_name_path)
    return tokenizer, model


def transformers_onnx_pipeline(model_name_path="distilgpt2"):
    tokenizer, model = download_from_transformers(model_name_path)
    test_torch_inference(model)
    onnx_path = generate_onnx_representation(model)
    test_onnx_inference(onnx_path, model.config)
    quantized_onnx_path = quantize(onnx_path)
    test_onnx_inference(quantized_onnx_path, model.config)


if __name__ == "__main__":
    transformers_onnx_pipeline("shibing624/code-autocomplete-distilgpt2-python")



"""
@author: Lowinli
@email: lowinli@outlook.com

"""

import os
import sys
import time
from transformers import AutoTokenizer, AutoModelForCausalLM
import torch
torch.set_num_threads(1)
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from fastgpt import (
    CausalLMModelForOnnxGeneration,
)

def evaluate_torch(max_length, num_beams):
    model = AutoModelForCausalLM.from_pretrained("distilgpt2")
    tokenizer = AutoTokenizer.from_pretrained("distilgpt2")
    
    prompt_text = "Natural language processing (NLP) is the ability of a computer program to understand human language as it is spoken and written"
    input_ids = tokenizer(
        prompt_text, return_tensors="pt", add_special_tokens=False
    ).input_ids
    start = time.time()
    for i in range(5):
        generated_ids = model.generate(
            input_ids,
            max_length=max_length + input_ids.shape[1],
            decoder_start_token_id=tokenizer.cls_token_id,
            eos_token_id=tokenizer.sep_token_id,
            output_scores=True,
            temperature=1,
            repetition_penalty=1.0,
            top_k=50,
            top_p=0.9,
            do_sample=True,
            num_return_sequences=1,
            length_penalty=2.0,
            early_stopping=True,
            num_beams=num_beams
        )
    waste_time = time.time() - start
    latency = round(waste_time / 5 * 1000, 3)
    print(tokenizer.decode(generated_ids[0], skip_special_tokens=True))
    print("=" * 20)
    return latency
    


def evaluate_fastgpt(max_length, num_beams):
    model = CausalLMModelForOnnxGeneration.from_pretrained("distilgpt2", threads=1)
    tokenizer = AutoTokenizer.from_pretrained("distilgpt2")

    prompt_text = "Natural language processing (NLP) is the ability of a computer program to understand human language as it is spoken and written"
    input_ids = tokenizer(
        prompt_text, return_tensors="pt", add_special_tokens=False
    ).input_ids
    start = time.time()
    for i in range(5):
        generated_ids = model.generate(
            input_ids,
            max_length=max_length + input_ids.shape[1],
            decoder_start_token_id=tokenizer.cls_token_id,
            eos_token_id=tokenizer.sep_token_id,
            output_scores=True,
            temperature=1,
            repetition_penalty=1.0,
            top_k=50,
            top_p=0.9,
            do_sample=True,
            num_return_sequences=1,
            length_penalty=2.0,
            early_stopping=True,
            num_beams=num_beams,
            use_cache=True
        )
    waste_time = time.time() - start
    latency = round(waste_time / 5 * 1000, 3)
    print(tokenizer.decode(generated_ids[0], skip_special_tokens=True))
    print("=" * 20)
    return latency


    
if __name__ == "__main__":
    with open("speed.md", "w") as f:
        f.write("## 生成速度评测(ms)\n\n")
        for max_length in [4, 8, 16, 32, 64]:
            f.write(f"+ 生成长度{max_length}评测\n")
            f.write("|模型框架|beam:1|beam:2|beam:3|beam:4|\n|-|-|-|-|-|\n|torch|")
            for num_beams in [1, 2, 3, 4]:
                latency = evaluate_torch(max_length, num_beams)
                print(f"torch: max_length{max_length}, num_beam{num_beams}, latency{latency}")
                f.write(f"{latency}|")
            f.write("\n|fastgpt|")
            for num_beams in [1, 2, 3, 4]:
                latency = evaluate_fastgpt(max_length, num_beams)
                print(f"fastgpt: max_length{max_length}, num_beam{num_beams}, latency{latency}")
                f.write(f"{latency}|")
            f.write("\n---\n")
            


import os
import torch
from transformers import PreTrainedModel, AutoConfig
from transformers.modeling_outputs import CausalLMOutputWithCrossAttentions
import onnxruntime
from typing import Optional, Tuple
import numpy as np
from .onnx_exporter import transformers_onnx_pipeline


class CausalLMModelForOnnxGeneration(PreTrainedModel):
    def __init__(
        self, onnx_model_path: str, model_path="", config=None, threads: int = 0
    ):
        if config is None:
            config = AutoConfig.from_pretrained(model_path)
        PreTrainedModel.__init__(self, config)
        # 加载onnxruntime
        sess_options = onnxruntime.SessionOptions()
        if threads:
            sess_options.intra_op_num_threads = threads
        self.session = onnxruntime.InferenceSession(onnx_model_path, sess_options)

    @classmethod
    def from_pretrained(cls, model_name_path: str, threads=0):
        onnx_path = os.path.join(model_name_path, "onnx/model-quantized.onnx")
        if not os.path.exists(onnx_path):
            transformers_onnx_pipeline(model_name_path)
        return cls(onnx_path, model_path=model_name_path, threads=threads)

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor]]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        if past_key_values is None:
            past_key_values_array = np.zeros(
                [
                    self.config.n_layer,
                    2,
                    input_ids.shape[0],
                    self.config.n_head,
                    0,
                    int(self.config.n_embd / self.config.n_head),
                ]
            ).astype(np.float32)
        else:
            past_key_values_array = (
                torch.stack([torch.stack(x) for x in past_key_values]).cpu().numpy()
            )
        if attention_mask is None:
            attention_mask = np.array(
                [[1] * int(past_key_values_array.shape[-2] + input_ids.shape[1])]*input_ids.shape[0]
            )
        else:
            attention_mask = attention_mask.cpu().numpy()
        ort_inputs = {
            "input_ids": input_ids.cpu().numpy(),
            "attention_mask": attention_mask,
            "past_key_values": past_key_values_array,
        }
        logits, past_key_values_array = self.session.run(None, ort_inputs)
        past_key_values = tuple(
            [tuple([torch.from_numpy(i) for i in x]) for x in past_key_values_array]
        )

        return CausalLMOutputWithCrossAttentions(
            loss=None,
            logits=torch.from_numpy(logits),
            past_key_values=past_key_values,
            hidden_states=None,
            attentions=None,
            cross_attentions=None,
        )

    @staticmethod
    def _reorder_cache(past: Tuple[Tuple[torch.Tensor]], beam_idx: torch.Tensor) -> Tuple[Tuple[torch.Tensor]]:
        """
        This function is used to re-order the `past_key_values` cache if [`~PreTrainedModel.beam_search`] or
        [`~PreTrainedModel.beam_sample`] is called. This is required to match `past_key_values` with the correct
        beam_idx at every generation step.
        """
        return tuple(
            tuple(past_state.index_select(0, beam_idx.to(past_state.device)) for past_state in layer_past)
            for layer_past in past
        )

    def prepare_inputs_for_generation(self, input_ids, past=None, **kwargs):
        token_type_ids = kwargs.get("token_type_ids", None)
        # only last token for inputs_ids if past is defined in kwargs
        if past:
            input_ids = input_ids[:, -1].unsqueeze(-1)
            if token_type_ids is not None:
                token_type_ids = token_type_ids[:, -1].unsqueeze(-1)

        attention_mask = kwargs.get("attention_mask", None)
        position_ids = kwargs.get("position_ids", None)

        if attention_mask is not None and position_ids is None:
            # create position_ids on the fly for batch generation
            position_ids = attention_mask.long().cumsum(-1) - 1
            position_ids.masked_fill_(attention_mask == 0, 1)
            if past:
                position_ids = position_ids[:, -1].unsqueeze(-1)
        else:
            position_ids = None

        return {
            "input_ids": input_ids,
            "past_key_values": past,
            "use_cache": kwargs.get("use_cache"),
            "position_ids": position_ids,
            "attention_mask": attention_mask,
            "token_type_ids": token_type_ids,
        }



"""
@author: Lowinli
@email: lowinli@outlook.com

"""

import torch


class ModelWrapper(torch.nn.Module):
    r"""
    这是一个onnx格式的装饰类，用于定义确定格式的矩阵输入和矩阵输出，从而做onnx格式转换
    """
    def __init__(self, model):
        super(ModelWrapper, self).__init__()
        self.model = model

    def forward(
        self,
        input_ids,
        attention_mask,
        past_key_values,
    ):
        res = self.model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            past_key_values=past_key_values,
            use_cache=True,
        )
        past_key_values_array = torch.stack(
            [torch.stack(x) for x in res.past_key_values]
        )  # 要保证wrapper的输出不能是python的列表，需要是torch矩阵，才能做onnx格式转换
        return res.logits, past_key_values_array



"""
@author: Lowinli
@email: lowinli@outlook.com

"""

import os
import sys
import unittest
from transformers import AutoTokenizer

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from fastgpt import (
    CausalLMModelForOnnxGeneration,
)


class Tests(unittest.TestCase):
    def test_model(self):
        model = CausalLMModelForOnnxGeneration.from_pretrained("distilgpt2")
        tokenizer = AutoTokenizer.from_pretrained("distilgpt2")

        prompt_text = "Natural language processing (NLP) is the ability of a computer program to understand human language as it is spoken and written"
        input_ids = tokenizer(
            prompt_text, return_tensors="pt", add_special_tokens=False
        ).input_ids
        generated_ids = model.generate(
            input_ids,
            max_length=64 + len(prompt_text),
            decoder_start_token_id=tokenizer.cls_token_id,
            eos_token_id=tokenizer.sep_token_id,
            output_scores=True,
            temperature=1,
            repetition_penalty=1.0,
            top_k=50,
            top_p=0.9,
            do_sample=True,
            num_return_sequences=1,
            length_penalty=2.0,
            early_stopping=True,
        )
        print(tokenizer.decode(generated_ids[0], skip_special_tokens=True))
        print("=" * 20)



from .model_wrapper import ModelWrapper
from .onnx_text_generate import CausalLMModelForOnnxGeneration
from .onnx_exporter import (
    generate_onnx_representation,
    quantize,
    test_torch_inference,
    test_onnx_inference,
    download_from_transformers,
    transformers_onnx_pipeline,
)


if __name__ == "__main__":
    generate_onnx_representation,
    quantize,
    test_torch_inference,
    test_onnx_inference,
    download_from_transformers,
    transformers_onnx_pipeline,
    ModelWrapper,
    CausalLMModelForOnnxGeneration



import setuptools
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="fastgpt",
    version="0.0.3",
    license="MIT License",
    author="LowinLi",
    author_email="lowinli@outlook.com",
    description="boost inference speed of GPT models in transformers by onnxruntime",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LowinLi/fastgpt",
    project_urls={
        "Repo": "https://github.com/LowinLi/fastgpt",
        "Bug Tracker": "https://github.com/LowinLi/fastgpt/issues",
    },
    keywords=[
        "GPT",
        "ONNX",
        "onnxruntime",
        "NLP",
        "model hub" "transformer",
        "quantization",
        "text generation",
        "summarization",
        "translation",
        "q&a",
        "qg",
        "machine learning",
        "fast inference",
        "CausalLM",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.5",
    install_requires=[
        "torch>=1.10.0",
        "onnx",
        "onnxruntime==1.10.0",
        "numpy>=1.22.2",
        "transformers>=4.19",
        "six==1.16.0"
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)
